﻿using Studer;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.IO.Ports;
using System.Linq;
using System.Text;

namespace StuderInterface.Helpers
{
    public class SerialComunication
    {
        /// <summary>
        /// lengt of receive buffer
        /// </summary>
        const int MAX_BUFF_LEN = 512;

        SerialPort serialPort;
        BackgroundWorker bwReceive;

        public SerialComunication()
        {
            bwReceive = new BackgroundWorker();
            bwReceive.WorkerReportsProgress = true;
            bwReceive.WorkerSupportsCancellation = true;
            bwReceive.DoWork += BwReceive_DoWork;
            bwReceive.ProgressChanged += BwReceive_ProgressChanged;
        }
        /// <summary>
        /// true when serial port is open
        /// </summary>
        public bool IsOpen { get { return serialPort != null ? serialPort.IsOpen : false; } }
        /// <summary>
        /// Use default configuration values of XCom232i:
        /// The default RS-232 configuration is defined as :
        /// A fixed baudrate of 38400 bps
        /// 1 start bit
        /// 8 bit of data, LSB first
        /// 1 parity bit
        /// even parity
        /// 1 stop bit
        /// </summary>
        /// <param name="portName"></param>
        /// <returns>null when ok, otherwise error mesage</returns>
        public string OpenPort(string portName)
        {
            try
            {
                serialPort = new SerialPort(portName);
                serialPort.BaudRate = 38400;
                serialPort.DataBits = 8;
                serialPort.Parity = Parity.Even;
                serialPort.StopBits = StopBits.One;
                serialPort.Handshake = Handshake.None;
                serialPort.Open();
                serialPort.WriteTimeout = 400;
                return null;
            }
            catch (Exception ex)
            {
                return ex.Message;
            }
        }
        /// <summary>
        /// close opened port
        /// </summary>
        /// <returns>null when Ok, otherwise error message</returns>
        public string ClosePort()
        {
            try
            {
                if (serialPort != null)
                {
                    serialPort.Close();
                    //serialPort.DataReceived -= SerialPort_DataReceived;
                    serialPort.Dispose();
                    serialPort = null;
                }

                return null;
            }
            catch (Exception ex)
            {
                return ex.Message;
            }
        }

        /// <summary>
        /// start background thread for repeated reading using ReadByte
        /// when correct studer frame is received it sends it.
        /// removes unreliability of onDataReceived
        /// </summary>
        /// <returns></returns>
        public bool StartReceiving()
        {
            if (!bwReceive.IsBusy && (serialPort != null) && serialPort.IsOpen)
            {
                bwReceive.RunWorkerAsync();
                return true;
            }
            return false;
        }
        /// <summary>
        /// stop background worker for reading
        /// </summary>
        /// <returns></returns>
        public bool StopReceiving()
        {
            if (bwReceive.IsBusy)
            {
                bwReceive.CancelAsync();
                return true;
            }

            return false;
        }
        /// <summary>
        /// send data
        /// </summary>
        /// <param name="data"></param>
        /// <returns></returns>
        public string SendData(IEnumerable<byte> data)
        {
            try
            {
                if ((serialPort != null) && serialPort.IsOpen && data != null && data.Count() > 0)
                {
                    serialPort.Write(data.ToArray(), 0, data.Count());
                    return null;
                }
                return "Port closed or no data to send";
            }
            catch(Exception ex)
            {
                return ex.Message;
            }
        }

        public delegate void FrameReceived(object sender, byte[] frameData);
        public FrameReceived OnFrameReceived;

        #region backgroundworker for receiving

        private void BwReceive_ProgressChanged(object sender, ProgressChangedEventArgs e)
        {
            if (e.ProgressPercentage == 1)
            {
                //received data
                byte[] data = e.UserState as byte[];
                OnFrameReceived?.Invoke(this, data);
            }
        }

        private void BwReceive_DoWork(object sender, DoWorkEventArgs e)
        {
            List<byte> rxBuffer = new List<byte>();//byte[MAX_BUFF_LEN];
            int dataLen = 0;
            bool readingData = false;
            while (!bwReceive.CancellationPending)
            {
                int i = serialPort.BytesToRead;
                if (i > 0)
                {
                    byte b = (byte)serialPort.ReadByte();

                    rxBuffer.Add(b);
                    //header 14 bytes, data, 2bytes of data check sum
                    if (readingData && (rxBuffer.Count >= (dataLen + 14 + 2)))
                    {
                        //data are complet, check check sum of data
                        if (FrameFnc.IsCorrectCheckSum(rxBuffer.Skip(14).ToArray()))
                        {
                            //we have corect frame
                            bwReceive.ReportProgress(1, rxBuffer.ToArray());

                        }
                        //reset receiving
                        rxBuffer.Clear();
                        readingData = false;
                    }

                    if (rxBuffer.Count >= 14 && !readingData)
                    {
                        //when we have more than 14 bytes (Frameheader lenght) in buffer chceck for start byte
                        int startPos = rxBuffer.IndexOf(0xAA);
                        if (startPos < 0)
                        {
                            //not start byte in buffer, reset buffer
                            rxBuffer.Clear();
                        }
                        else
                        {
                            //remove bytes before start byte
                            if (startPos > 0)
                                rxBuffer.RemoveRange(0, startPos);
                        }

                        if (rxBuffer.Count >= 14)
                        {
                            //start character presented and there is possible to decode frame header
                            //check check sum - ignore start byte from check sum !!!!
                            var header = rxBuffer.Take(14).ToArray();
                            if (FrameFnc.IsCorrectCheckSum(header.Skip(1).ToArray()))
                            {
                                //read how long is data part of frame
                                dataLen = FrameFnc.GetDataLen(header);
                                readingData = true;
                            }
                            else
                            {
                                //not correct header remove startCharacter
                                rxBuffer.RemoveAt(0);
                            }
                        }
                    }

                    if (rxBuffer.Count > MAX_BUFF_LEN)
                    {
                        rxBuffer.RemoveAt(0);
                    }
                }
                else
                {
                    System.Threading.Thread.Sleep(50);
                }
            }
        }
        #endregion
    }
}
