﻿using Studer;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Studer
{
    public class Frame
    {
        /// <summary>
        /// service bytes
        /// service_flags 1 byte
        /// service_id  1 byte
        /// </summary>
        List<byte> dataBytes;
        byte[] checkSumBytes = new byte[2];

        #region constructors
        public Frame()
        {
            Header = new FrameHeader();
            dataBytes = new List<byte>() { 0x0, 0x1 };
        }

        public Frame(IEnumerable<byte> bytedata)
        {
            if (bytedata != null && bytedata.Count() >= 14)
                Header = new FrameHeader(bytedata.Take(14));
            else
                Header = new FrameHeader();

            if (bytedata != null && bytedata.Count() >= 16)
            {
                dataBytes = new List<byte>();
                var len = bytedata.Count();
                //data bytes without check sum
                dataBytes.AddRange(bytedata.Skip(14).Take(len -16));
                checkSumBytes = bytedata.Skip(len - 2).ToArray();
            }
            else
                dataBytes = new List<byte>() { 0x0, 0x1 };
        }
        #endregion

        #region properties
        public FrameHeader Header { get; private set; }
        /// <summary>
        /// error flag, 0 in case of success, 1 if an error occurred. In case of a request, error is always 0.
        /// </summary>
        public bool ServiceFlag0 { get { return  ((byte)dataBytes[0] & 0x1) != 0; } }
        /// <summary>
        /// is_response flag, 0 if it is a request from the DTE to the Xcom-232i, 1 if it is response from the Xcom-232i.
        /// </summary>
        public bool IsResponse 
        {
            get { return ((byte)dataBytes[0] & 0x2) != 0; }
            set
            {
                if (value) //set bit 1 in byte
                    dataBytes[0] = (byte)(dataBytes[0] | 0x2);
                else //clear bit 1 in byte
                    dataBytes[0] = (byte)(dataBytes[0] & 0xBF);
            }
        }
        /// <summary>
        /// 
        /// </summary>
        public bool IsError
        {
            get { return ServiceFlag0; }
        }
        /// <summary>
        /// READ_PROPERTY (0x1)
        /// WRITE_PROPERTY (0x2)
        /// </summary>
        public ServiceIdEnum ServiceId
        {
            get { return dataBytes[1] == 0x2 ? ServiceIdEnum.WRITE_PROPERTY: ServiceIdEnum.READ_PROPERTY; }
            set
            {
                dataBytes[1] = (byte)value;
            }
        }
        /// <summary>
        /// two bytes 
        /// dataBytes[2]..dataBytes[3]
        /// When not correct data in bytes returns undefined
        /// </summary>
        public ObjectTypeEnum ObjectType
        {
            get
            {
                var val = FrameFnc.Convert2BytesToUInt16(dataBytes.Skip(2).Take(2).ToArray());
                try
                {
                    return (ObjectTypeEnum)val;
                }
                catch
                {
                    return ObjectTypeEnum.UNDEFINED;
                }
            }
            set
            {
                var buf = FrameFnc.ConvertUInt16to2bytes((UInt16)value);
                while (dataBytes.Count < 4)
                    dataBytes.Add(0x0);

                dataBytes[2] = buf[0];
                dataBytes[3] = buf[1];
            }
        }

       

        /// <summary>
        /// each object type has its own object ids
        /// 4 bytes
        /// dataBytes[4]..dataBytes[7]
        /// </summary>
        public UInt32 ObjectId
        {
            get { return FrameFnc.Convert4ByteToUInt32(dataBytes.Skip(4).Take(4)); }
            set
            {
                var buf = FrameFnc.ConvertUInt32to4bytes(value);
                while (dataBytes.Count < 8)
                    dataBytes.Add(0x0);

                for (int i = 0; i < 4; i++)
                {
                    dataBytes[4 + i] = buf[i];
                }
            }
        }
        /// <summary>
        /// each ObjectId has its own set of PropertyIds'.
        /// 2 bytes
        /// dataBytes[8]..dataBytes[9]
        /// </summary>
        public UInt16 PropertyId
        {
            get { return FrameFnc.Convert2BytesToUInt16(dataBytes.Skip(8).Take(2).ToArray()); }
            set
            {
                var buf = FrameFnc.ConvertUInt16to2bytes(value);
                while (dataBytes.Count < 10)
                    dataBytes.Add(0x0);

                dataBytes[8] = buf[0];
                dataBytes[9] = buf[1];
            }
        }
        /// <summary>
        /// property data starts at dataBytes[10] .. variable len
        /// </summary>
        public IEnumerable<byte> PropertyData
        {
            get 
            {
                var len = dataBytes.Count;
                return dataBytes.Skip(10); 
            }
            set
            {
                if (dataBytes.Count > 10)
                    dataBytes.RemoveRange(10, dataBytes.Count - 10);

                dataBytes.AddRange(value);
            }
        }
        /// <summary>
        /// 0 - in case of error
        /// </summary>
        public UInt16 GetDataCheckSumAsUInt
        {
            get
            {
                int a = checkSumBytes[0];
                int b = checkSumBytes[1];
                return (UInt16)(a * 256 + b);
            }
        }
        /// <summary>
        /// calculate check sum of data in buffer .
        /// 0 is usually error
        /// </summary>
        public UInt16 CalculatedCheckSumAsUInt
        {
            get
            {
                if (dataBytes != null)
                {
                    return FrameFnc.CalculateCheckSum(dataBytes.ToArray());
                }
                else
                    return 0;
            }
        }
        #endregion

        #region methods
        public void SetFloatAsPropertyData(float value)
        {
            var bytes = FrameFnc.FloatTo4Bytes(value);
            if (dataBytes.Count > 10)
                dataBytes.RemoveRange(10, dataBytes.Count - 10);

            dataBytes.AddRange(bytes);
        }
        public void SetUInt16AsPropertyData(UInt16 value)
        {
            var bytes = FrameFnc.ConvertUInt16to2bytes(value);
            if (dataBytes.Count > 10)
                dataBytes.RemoveRange(10, dataBytes.Count - 10);

            dataBytes.AddRange(bytes);
        }
        /// <summary>
        /// calculates length of data, sets this value in header, calculates chceck sum of header and data frame. 
        /// </summary>
        /// <returns></returns>
        public IEnumerable<byte> GetBytes()
        {
            int dataLen = dataBytes.Count;
            Header.DataLen = dataLen;
            List<byte> result = Header.GetFrameHeaderBytes.ToList();
            //append dataframe
            result.AddRange(dataBytes);
            //append datframe checksum
            var dataCS = FrameFnc.CalculateCheckSumBytes(dataBytes);
            result.AddRange(dataCS);
            return result;
        }
        #endregion
    }
}
