﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Studer
{
    /// <summary>
    /// methods for working with frame byte array
    /// </summary>
    public static class FrameFnc
    {
        #region public functions
        /// <summary>
        /// calculate studer like check sum for given byte span
        /// </summary>
        /// <param name="byteEnum">byte enum on which is calculated check sum</param>
        /// <returns></returns>
        public static byte[] CalculateCheckSumBytes(IEnumerable<byte> byteEnum)
        {
            int a = 0xFF;
            int b = 0;
            if (byteEnum != null)
            {
                for (int i = 0; i < byteEnum.Count(); i++)
                {
                    //add value and keep modulo of 256 - that was original
                    a = (a + byteEnum.ElementAt(i)) % 0x100; // mod 0x100; 
                    b = (b + a) % 0x100; //mod 0x100;
                }
            }
            return new byte[] { (byte)a, (byte)b };
        }
        /// <summary>
        /// calculate studer like check sum for given byte span
        /// </summary>
        /// <param name="byteSpan">byte span on which is calculated check sum</param>
        /// <returns></returns>
        public static byte[] CalculateCheckSumBytes(ReadOnlySpan<byte> byteSpan)
        {
            int a = 0xFF;
            int b = 0;
            if (byteSpan != null)
            {
                for (int i = 0; i < byteSpan.Length; i++)
                {
                    //add value and keep modulo of 256 - that was original
                    a = (a + byteSpan[i]) % 0x100; // mod 0x100; 
                    b = (b + a) % 0x100; //mod 0x100;
                }
            }
            return new byte[] { (byte)a, (byte)b };
        }
        /// <summary>
        /// calculate studer like check sum for given byte span
        /// </summary>
        /// <param name="byteSpan">byte span on which is calculated check sum</param>
        /// <returns></returns>
        public static UInt16 CalculateCheckSum(ReadOnlySpan<byte> byteSpan)
        {
            int a = 0xFF;
            int b = 0;
            if (byteSpan != null)
            {
                for (int i = 0; i < byteSpan.Length; i++)
                {
                    //add value and keep modulo of 256 - that was original
                    a = (a + byteSpan[i]) % 0x100; // mod 0x100; 
                    b = (b + a) % 0x100; //mod 0x100;
                }
            }
            return (UInt16)((a * 256) + b);
        }
        /// <summary>
        /// converts two bytes AB to uint value (A*256 + B)
        /// </summary>
        /// <param name="a"></param>
        /// <param name="b"></param>
        /// <returns></returns>
        public static UInt16 GetCheckSum(byte a, byte b)
        {
            return (UInt16)(a * 256 + b);
        }
        /// <summary>
        /// return true when byte array is correct frame header or data.
        /// Check sum is taken from last two positions of byteArr
        /// </summary>
        /// <param name="byteSpan"></param>
        /// <returns></returns>
        public static bool IsCorrectCheckSum(ReadOnlySpan<byte> byteSpan)
        {
            if (byteSpan != null && byteSpan.Length > 2)
            {
                var len = byteSpan.Length;
                var cs = GetCheckSum(byteSpan[len - 2], byteSpan[len - 1]);

                var calccs = CalculateCheckSum(byteSpan.Slice(0, len - 2));
                if (cs == calccs)
                    return true;
            }
            return false;
        }

        /// <summary>
        /// 2bytes converted to int (little endian) from header bytes
        /// byte[10] + byte[11]*256
        /// </summary>
        public static int GetDataLen(ReadOnlySpan<byte> headerByteSpan)
        {
            if (headerByteSpan != null && headerByteSpan.Length > 11)
            {
                return headerByteSpan[10]  + headerByteSpan[11] * 256;
            }
            else
                return 0;
        }
        /// <summary>
        /// 4 bytes converted to UInt32 (little endian)
        /// byte[0] + byte[1]*256 + byte[2]*65536 + byte[3]*65536*256
        /// </summary>
        /// <param name="fourbyte"></param>
        /// <returns></returns>
        public static UInt32 Convert4ByteToUInt32(ReadOnlySpan<byte> fourbyte)
        {
            if (fourbyte != null && fourbyte.Length >= 4)
            {
                if (!BitConverter.IsLittleEndian)
                {
                    var reversed = fourbyte.ToArray().Reverse().ToArray();
                    return BitConverter.ToUInt32(reversed);
                }
                else
                    return BitConverter.ToUInt32(fourbyte);
            }
            else
                return 0;
        }
        /// <summary>
        /// 4 bytes converted to UInt32 (little endian)
        /// byte[0] + byte[1]*256 + byte[2]*65536 + byte[3]*65536*256
        /// </summary>
        /// <param name="fourbyte"></param>
        /// <returns></returns>
        public static UInt32 Convert4ByteToUInt32(IEnumerable<byte> fourbyte)
        {
            if (fourbyte != null && fourbyte.Count() >= 4)
            {
                if (!BitConverter.IsLittleEndian)
                {
                    var reversed = fourbyte.Take(4).ToArray().Reverse().ToArray();
                    return BitConverter.ToUInt32(reversed);
                }
                else
                    return BitConverter.ToUInt32(fourbyte.Take(4).ToArray());
            }
            else
                return 0;
        }
        /// <summary>
        /// converts UInt32 value to 4 bytes using little endian encoding
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static byte[] ConvertUInt32to4bytes(UInt32 value)
        {
            if (BitConverter.IsLittleEndian)
            {
                return BitConverter.GetBytes(value);
            }
            else
            {
                var arr = BitConverter.GetBytes(value);
                return arr.Reverse().ToArray();
            }
        }
        /// <summary>
        /// converts UInt16 value to 2 bytes using little endian encoding
        /// </summary>
        /// <param name="value"></param>
        /// <returns></returns>
        public static byte[] ConvertUInt16to2bytes(UInt16 value)
        {
            if (BitConverter.IsLittleEndian)
            {
                return BitConverter.GetBytes(value);
            }
            else
            {
                var arr = BitConverter.GetBytes(value);
                return arr.Reverse().ToArray();
            }
        }
        /// <summary>
        /// convertw two bytes to UInt16
        /// </summary>
        /// <param name="twobyte"></param>
        /// <returns></returns>
        public static UInt16 Convert2BytesToUInt16(ReadOnlySpan<byte> twobyte)
        {
            if (twobyte != null && twobyte.Length >= 2)
            {
                return (UInt16)(twobyte[0] + twobyte[1] * 256);
            }
            return 0;
        }
        /// <summary>
        /// convert float value to 4 bytes (little endian DCBA)
        /// </summary>
        /// <param name="floatValue"></param>
        /// <returns></returns>
        public static byte[] FloatTo4Bytes(float floatValue)
        {
            if (BitConverter.IsLittleEndian)
            {
                return BitConverter.GetBytes(floatValue);
            }
            else
            {
                var arr = BitConverter.GetBytes(floatValue);
                return arr.Reverse().ToArray();
            }
        }
        #endregion
    }
}
