﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace Studer
{
    /// <summary>
    /// length of 14 bytes including start byte 0xAA and two CRC bytes at the end
    /// </summary>
    public class FrameHeader
    {
        byte[] hdata;
        public FrameHeader()
        {
            hdata = new byte[14];
            //starting byte
            hdata[0] = 0xAA;
        }
        /// <summary>
        /// constructs object with filling inernal buffer with bytes.
        /// 
        /// </summary>
        /// <param name="bytedata">14 bytes, including start byte and two check sum bytes</param>
        public FrameHeader(IEnumerable<byte> bytedata)
        {
            if (bytedata != null && bytedata.Count() == 14)
            {
                hdata = bytedata.ToArray();
            }
        }

        #region properties
        /// <summary>
        /// always 0xAA
        /// </summary>
        public byte StartByte { get { return hdata[0]; } }

        public byte FrameFlag { get { return hdata[1]; } set { hdata[1] = value; } }
        /// <summary>
        /// 4bytes converted to UInt32 (little endian)
        /// hdata[2]...hdata[5]
        /// </summary>
        public UInt32 SrcAddr 
        { 
            get { return FrameFnc.Convert4ByteToUInt32(hdata.AsSpan(2,4)); }
            set 
            {
                var buff = FrameFnc.ConvertUInt32to4bytes(value);
                buff.CopyTo(hdata, 2); 
            } 
        }

        /// <summary>
        /// 4bytes converted to Int32 (little endian)
        /// hdata[6]...hdata[9]
        /// </summary>
        public UInt32 DstAddr 
        {
            get { return FrameFnc.Convert4ByteToUInt32(hdata.AsSpan(6, 4)); }
            set
            {
                var buff = FrameFnc.ConvertUInt32to4bytes(value);
                buff.CopyTo(hdata, 6);
            }
        }

        /// <summary>
        /// 2bytes converted to int (little endian)
        /// hdata[10]...hdata[11]
        /// </summary>
        public int DataLen 
        {
            get { return FrameFnc.GetDataLen(hdata); }
            set 
            {
                var buff = FrameFnc.ConvertUInt16to2bytes((UInt16)value);
                buff.CopyTo(hdata, 10);
            } 
        }
        
        /// <summary>
        /// 2bytes from position 12, 13, converted to uint
        /// uses byte order  
        /// Buffer[12] * 256 + Buffer[13]
        /// </summary>
        public uint GetHeaderCheckSumAsUInt
        {
            get
            {
                if (hdata.Length >= 14)
                {
                    int a = hdata[12];
                    int b = hdata[13];
                    return (uint)( a*256 + b);
                }
                else
                    return 0;
            }
        }
        /// <summary>
        /// calculate check sum from first 12 bytes of data in buffer.
        /// Skips starting byte !!!
        /// 0 is usually error - short buffer
        /// </summary>
        public uint CalculatedCheckSumAsUInt
        {
            get
            {
                if (hdata.Length >= 12)
                {
                    //used
                    return FrameFnc.CalculateCheckSum(hdata.AsSpan().Slice(1,11));
                }
                else
                    return 0;
            }
        }
        /// <summary>
        /// return 14 bytes of header including valid header checksum 
        /// </summary>
        public byte[] GetFrameHeaderBytes
        {
            get
            {
                byte[] checkSum = FrameFnc.CalculateCheckSumBytes(hdata.AsSpan().Slice(1, 11));
                byte[] result = new byte[14];
                hdata.CopyTo(result, 0);
                checkSum.CopyTo(result, 12);
                return result;
            }
        }

        #endregion

        #region methods
        #endregion
    }
}
