﻿using Studer;
using StuderInterface.Helpers;
using StuderInterface.Settings;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Text;
using System.Threading;

namespace StuderInterface
{
    /// <summary>
    /// sends and receives data to studer targets
    /// </summary>
    public class StuderComunicator
    {
        /// <summary>
        /// time out for waiting for response 2s
        /// </summary>
        const int timeout = 2000;
        /// <summary>
        /// minimal time span between command 300ms
        /// </summary>
        const int MinCommandTimeSpan = 300;
        /// <summary>
        /// number of defined priorities
        /// </summary>
        const int NumberOfPriorities = 3;
        /// <summary>
        /// priority 1 is read each 10s, 2 each 60s
        /// </summary>
        int[] PriorityPollingTimes = new int[NumberOfPriorities] { 0, 10, 60 };

        SISettings settings;
        BackgroundWorker bw;
        
        Action<string> writeLog;
        public StuderComunicator(SISettings settings, Action<string> writeLog)
        {
            this.settings = settings;
            this.writeLog = writeLog;
            bw = new BackgroundWorker();
            bw.DoWork += Bw_DoWork;
            bw.RunWorkerCompleted += Bw_RunWorkerCompleted;
            bw.ProgressChanged += Bw_ProgressCanged;
            bw.WorkerReportsProgress = true;
            bw.WorkerSupportsCancellation = true;
            //correct priority in settings
            if (this.settings != null)
            {
                foreach (var par in this.settings.Parameters)
                    if (par.Priority >= NumberOfPriorities)
                        par.Priority = NumberOfPriorities - 1;
            }
        }
        /// <summary>
        /// true when is running
        /// </summary>
        public bool IsRunning { get { return bw.IsBusy; } }


        public bool Start()
        {
            if (bw.IsBusy)
                return false;

            // open communication port to XCom232
            bw.RunWorkerAsync();
            return true;
        }

        public void Stop()
        {
            if (bw.IsBusy)
                bw.CancelAsync();
        }

        public delegate void DataReceived(object sender, object data);
        public DataReceived OnDataReceived;


        bool gotResponse = false;
        object response = null;

        private void Bw_DoWork(object sender, DoWorkEventArgs e)
        {
            try
            {
                SerialComunication comunication = new SerialComunication();
                comunication.OnFrameReceived += OnFrameReceived;
                string err = comunication.OpenPort(settings.PortName);
                if (!string.IsNullOrEmpty(err))
                {
                    //error sending data
                    bw.ReportProgress(0, new ErrResponse(null, err)); //error
                    return;
                }
                comunication.StartReceiving();
                //requests to send
                int cmdCount = settings.Parameters != null ? settings.Parameters.Length:0;
                if (cmdCount > 0)
                {
                    // priority 0 ... 5
                    List<int> processPriorities = new List<int>() { 0 };
                    int priorityMax = settings.Parameters.Max(a => a.Priority);
                    int idx = 0;
                    //when we start to wait for response
                    long startTicks;
                    //when command was send
                    long startCommandTicks;
                    //for priority 1, 2 
                    long[] priorityStartTicks = new long[] { DateTime.Now.Ticks, DateTime.Now.Ticks, DateTime.Now.Ticks };
                    //start infinity loop
                    while (!bw.CancellationPending)
                    {
                        
                        var reqItem =  settings.Parameters[idx];
                        if (processPriorities.Contains( reqItem.Priority ))
                        {
                            //process in this round, build request frame 
                            Frame frame = new Frame();
                            frame.Header.SrcAddr = StuderConst.ADDR_RESPONSE;
                            frame.Header.DstAddr = (uint)reqItem.Addr;
                            frame.Header.FrameFlag = 0;
                            frame.IsResponse = false;
                            frame.ServiceId = ServiceIdEnum.READ_PROPERTY;
                            frame.ObjectType = (ObjectTypeEnum)reqItem.ObjType;
                            frame.ObjectId = (uint)reqItem.ObjectId;
                            frame.PropertyId = (ushort)reqItem.PropertyId;
                            var req = frame.GetBytes();
                            gotResponse = false;
                            //note when we start sending command
                            startCommandTicks = DateTime.Now.Ticks;
                            err = comunication.SendData(req);
                            if (!string.IsNullOrEmpty(err))
                            {
                                //error sending data
                                bw.ReportProgress(0, new ErrResponse(reqItem, err)); //error
                            }
                            else
                            {
                                startTicks = DateTime.Now.Ticks;
                                //check for response until getting response or timeout
                                while (!gotResponse && (new TimeSpan(DateTime.Now.Ticks - startTicks).TotalMilliseconds < timeout))
                                {
                                    Thread.Sleep(20);
                                }

                                if (!gotResponse)
                                {
                                    //no response during timeout interval - report as timeout
                                    bw.ReportProgress(0, new ErrResponse(reqItem, "timeout") ); //timeout
                                }
                                else
                                {
                                    //we got response, report it
                                    bw.ReportProgress(1, response);
                                }
                            }
                            //let MinCommandTimeSpan time span between commands
                            int runSpan = Convert.ToInt32( new TimeSpan(DateTime.Now.Ticks - startCommandTicks).TotalMilliseconds );
                            if (runSpan < MinCommandTimeSpan)
                                Thread.Sleep((MinCommandTimeSpan - runSpan));
                        }

                        //next command
                        idx++;
                        if (idx >= cmdCount)
                        {
                            //reset priority list
                            processPriorities.Clear(); processPriorities.Add(0);
                            //set priority, when round is finished and is time to read lowest priorities
                            long actualTicks = DateTime.Now.Ticks;
                            for (int i = priorityMax; i > 0; i--)
                            {
                                if (new TimeSpan(actualTicks - priorityStartTicks[i]).Milliseconds > PriorityPollingTimes[i])
                                {
                                    processPriorities.Add(i);
                                    priorityStartTicks[i] = actualTicks;
                                }
                            }
                            //from beginning
                            idx = 0;
                        }
                    }

                }
                comunication.StopReceiving();
            }
            finally
            {

            }
        }

        private void Bw_RunWorkerCompleted(object sender, RunWorkerCompletedEventArgs e)
        {
            if (e.Error != null)
            {
                if (writeLog != null)
                    writeLog(e.Error.Message);
            }
            else
            if (writeLog != null)
                writeLog("Comunication with XCom232 finished");
        }


        private void Bw_ProgressCanged(object sender, ProgressChangedEventArgs e)
        {
            if (e.ProgressPercentage == 0)
            {
                OnDataReceived?.Invoke(this, e.UserState as ErrResponse);
            }
            else
            if (e.ProgressPercentage == 1)
            {
                //send received frame
                OnDataReceived?.Invoke(this, e.UserState as Frame);
            }
        }

        private void OnFrameReceived(object sender, byte[] frameData)
        {
            ;
            if (frameData != null)
            {
                //FrameBytesHex = string.Join(" ", frameData.Select(a => a.ToString("X2")).ToArray());

                //build frame
                Frame frame = new Frame(frameData);
                //check dest addres
                if (frame.Header.DstAddr == StuderConst.ADDR_RESPONSE)
                {
                    gotResponse = true;
                    response = frame;
                }
            }
        }
    }

    public class ErrResponse
    {
        public ErrResponse(ReqItem reqItem, string errMsg)
        {
            this.ReqItem = new ReqItem( reqItem);
            this.ErrMsg = errMsg;
        }
        public ReqItem ReqItem { get; set; }
        public string ErrMsg { get; set; }

    }
}
