﻿using StuderEmulator.Studer;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;

namespace StuderInterface.Studer
{
    public static class StuderConverter
    {
        public static object ValueToCsObject(FormatEnum stformat, IEnumerable<byte> data)
        {
            if (data == null || data.Count() == 0)
                return null;

            switch (stformat)
            {
                case FormatEnum.SCOM_FORMAT_BOOL:
                    //1 byte
                    return (bool)(data.ElementAt(0) == (byte)1);
                // 2 bytes - 16bit uinteger
                case FormatEnum.SCOM_FORMAT_SHORT_ENUM:
                case FormatEnum.SCOM_FORMAT_ERROR:
                    return ToUInt16(data);
                // 4 bytes - 32 bit signed integer
                case FormatEnum.SCOM_FORMAT_INT32:
                    return ToInt32(data);
                // 4 bytes - 32 bit unsiged integer
                case FormatEnum.SCOM_FORMAT_LONG_ENUM:
                    return ToUInt32(data);
                // 4 bytes float
                case FormatEnum.SCOM_FORMAT_FLOAT:
                    return (float)ToFloatSafe(data);
                //n- bytes ISO_8859-15 string of 8 bit characters
                case FormatEnum.SCOM_FORMAT_STRING:
                    var encoding = Encoding.GetEncoding(28605);
                    return encoding.GetString(data.ToArray()) as string;
                case FormatEnum.SCOM_FORMAT_BYTE_STREAM:
                    return data.ToArray();
            }

            return null;
        }
        /// <summary>
        /// 2 bytes - little endian encoding
        /// </summary>
        /// <param name="twobyte"></param>
        /// <returns></returns>
        public static UInt16 ToUInt16(IEnumerable<byte> twobyte)
        {
            if (twobyte != null && twobyte.Count() >= 2)
            {
                return (UInt16)(twobyte.ElementAt(0) + twobyte.ElementAt(1) * 256);
            }
            return 0;
        }
        /// <summary>
        /// 4 bytes - little endian encoding
        /// </summary>
        /// <param name="fourbyte"></param>
        /// <returns></returns>
        public static Int32 ToInt32(IEnumerable<byte> fourbyte)
        {
            Int32 fb = 0;
            if (fourbyte != null && fourbyte.Count() >= 4)
            {

                //higher is most right byte
                for (int i = 3; i >= 0; i--)
                {
                    fb = fb + fourbyte.ElementAt(i);
                    if (i > 0)
                        fb = fb << 8;
                }
            }
            return fb;
        }
        /// <summary>
        /// 4 bytes - little endian encoding
        /// </summary>
        /// <param name="fourbyte"></param>
        /// <returns></returns>
        public static UInt32 ToUInt32(IEnumerable<byte> fourbyte)
        {
            UInt32 fb = 0;
            if (fourbyte != null && fourbyte.Count() >= 4)
            {
                
                //higher is most right byte
                for (int i = 3; i >= 0; i--)
                {
                    fb = fb + fourbyte.ElementAt(i);
                    if (i > 0)
                        fb = fb << 8;
                }
            }
            return fb;
        }

        /// <summary>
        /// 4bytes to float value
        /// IEE754
        /// litle endian
        /// </summary>
        /// <param name="f"></param>
        /// <returns></returns>
        public static float ToFloatSafe(IEnumerable<byte> fourByte)
        {
            if (fourByte == null || fourByte.Count() < 4)
                return 0;

            uint fb = 0;
            //higher is most right byte
            for (int i = 3; i >= 0; i--)
            {
                fb = fb + fourByte.ElementAt(i);
                if (i > 0)
                    fb = fb << 8;
            }

            int sign = (int)((fb >> 31) & 1);
            int exponent = (int)((fb >> 23) & 0xFF);
            int mantissa = (int)(fb & 0x7FFFFF);

            float fMantissa;
            float fSign = sign == 0 ? 1.0f : -1.0f;

            if (exponent != 0)
            {
                exponent -= 127;
                fMantissa = 1.0f + (mantissa / (float)0x800000);
            }
            else
            {
                if (mantissa != 0)
                {
                    // denormal
                    exponent -= 126;
                    fMantissa = 1.0f / (float)0x800000;
                }
                else
                {
                    // +0 and -0 cases
                    fMantissa = 0;
                }
            }

            float fExponent = (float)Math.Pow(2.0, exponent);
            float ret = fSign * fMantissa * fExponent;
            return ret;
        }
    }

}
